-- Copyright  1995-1997 TeCGraf, PUC-Rio. 
-- Written by Renato Ferreira Borges, Roberto Ierusalimschy and
-- Anna Magdalena Hester
-- All rights reserved. 
-- 
-- Permission is hereby granted, without written agreement and
-- without license or royalty fees, to use, copy, modify, and distribute
-- this software and its documentation for any purpose, subject to the
-- following conditions: 
-- 
-- The above copyright notice and this permission notice shall appear in
-- all copies or substantial portions of this software. 
-- 
-- The names "CGI Lua" and "HTML Lua" cannot be used for any modified form of
-- this software that does not originate from the authors.
-- 
-- The authors specifically disclaim any warranties, including, but not
-- limited to, the implied warranties of merchantability and fitness for a
-- particular purpose. The software provided hereunder is on an "as is"
-- basis, and the authors have no obligation to provide maintenance,
-- support, updates, enhancements, or modifications. In no event shall
-- TeCGraf, PUC-Rio, or the authors be liable to any party for direct,
-- indirect, special, incidental, or consequential damages arising out of
-- the use of this software and its documentation. 

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------


-- creates the table cgilua, used as a namespace

cgilua = {}


-- moves the global symbols to the table <cgilua>

cgilua.error = cgilua_error
cgilua_error = nil

cgilua.conffile = cgilua_conffile
cgilua_conffile = nil

cgilua.param = cgilua_param
cgilua_param = nil

cgilua.userfallback = cgilua_userfallback
cgilua_userfallback = nil


-- <cgilua.dofile> is used to 'do' a cgilua file,
-- generating an adequate error message in case of error

function cgilua.dofile( file )
  if not dofile( cgilua.directory..file ) then
    cgilua.error( format("Cannot open CGILua file <CODE>%s</CODE>, check installation.", file) )
  end
end


-- <cgilua.directory> points to the cgilua script directory.

cgilua.directory = gsub( cgilua.conffile, "[^\\/:]*$", "" )


-- <cgilua.external>, <cgilua.restore_external> and <cgilua.remove_external>
-- are used to remove some external symbols from the environment
-- during the execution of the 'map.lua' file

cgilua.external = { "readfrom", "writeto", "appendto", "execute",
                    "remove", "rename", "loadlibrary" }

function cgilua.restore_external()
  local i = 1
  while cgilua.external[i] do
    local elem = cgilua.external[i]
    if cgilua.external[elem] then
      ref = getrestricted( cgilua.external[elem] )
      dostring( elem .. " = ref" )
    end
    i = i + 1
  end
end

function cgilua.remove_external()
  local i = 1
  while cgilua.external[i] do
    local elem = cgilua.external[i]
    local func = dostring( "return "..elem )
    if func then
      cgilua.external[elem] =
        restrict( func, cgilua.restore_external )
      dostring( elem .. "= nil" ) 
    end
    i = i + 1
  end
end


-- the function <cgilua.mappathinfo> maps the environment variable
-- PATH_INFO to the physical path of the lua file to be executed (the script).

function cgilua.mappathinfo()

  local path_info = getenv("PATH_INFO")

  -- requires PATH_INFO
  if not path_info or path_info == '' or path_info == '/' then
    cgilua.error( "No CGILua script specified in url." );
  end

  -- do not allow paths with '..'
  if strfind( path_info, "%.%." ) then
    cgilua.error( "Invalid lua file (`..' not allowed)." )
  end

  -- checks if PATH_INFO is in the format ~user/path
  local luafile
  local i, f, user, path = strfind( path_info, "^/~([A-Za-z0-9_-]*)(/.*)$" )

  if not i then

    -- if the url does not specifies an user, 
    -- take the script from the directory <cgilua.root_directory>

    luafile = cgilua.root_directory .. path_info

  else

    -- if there's a map (user -> diretory) in the table <cgilua.users>,
    -- take the script from this directory
    -- if not, use PATH_TRANSLATED changing the sub-path
    -- from <cgilua.html_directory> to <cgilua.lua_directory>

    if cgilua.users[user] then
      luafile = cgilua.users[user]..path
    else
      local path_translated = getenv("PATH_TRANSLATED")
      if path_translated then
        luafile = gsub( path_translated,
                        cgilua.html_directory,
                        cgilua.lua_directory )
      else
        cgilua.error( format("User '%s' unknown.", user ) )
      end
    end

  end

  return luafile
end


-- the function <cgilua.getuserscript> finds the script file to be executed.

function cgilua.getuserscript()
  if cgilua.param then
    return cgilua.param
  else
    return cgilua.mappathinfo()
  end
end


-- define the default values for the variables expected to be defined
-- in the file 'map.lua'

-- the following four variables are used by the function
-- <cgilua.mappathinfo> and controls how the 
-- environment variable PATH_INFO is mapped to the
-- physical path of the lua script to be executed.
-- (note: it's only necessary to set them if cgilua is referenced directly
-- in the url, using PATH_INFO to specify the lua script.)
-- the variable <cgilua.users> is a table mapping
-- from users to physical directories.
-- if PATH_INFO is in the format ~user/path
-- and there's an entry in the table <cgilua.users>
-- for the user specified in PATH_INFO,
-- the physical path will be PATH_INFO with <cgilua.users[user]>
-- in place of ~user.
-- if there's no map, the physical path will be the environment 
-- variable PATH_TRANSLATED with <cgilua.lua_directory>
-- in place of <cgilua.html_directory>.
-- note: some httpd servers do not set the environment
-- variable PATH_TRANSLATED. in this case, there must be an entry
-- in the table <cgilua.users> for each user to be used in PATH_INFO.
-- if PATH_INFO is not in the format ~user/path, the physical path will be 
-- <cgilua.root_directory> concatenated with PATH_INFO.
-- if the preprocess facility is turned on 
-- <cgilua.filename_filter> contain the pattern matching expression
-- for filename filtering. the files which match the filter are
-- preprocessed.
-- the decoded form data are put in the table <cgi>.
-- for each form element the function <insertfield> 
-- is called with the pair <name>,<value>

cgilua.root_directory = cgilua.directory .. 'www/'
cgilua.lua_directory  = "/public_html/"
cgilua.html_directory = "/public_html/"
cgilua.users = { }
cgilua.filename_filter = ".*%.html?$"

cgi = {}

function cgilua.insertfield( name, value )
  if not cgi[name] then
    cgi[name] = value
  else
    cgi[name] = cgi[name]..','..value
  end
end



-- <cgilua.domap> executes the file 'map.lua', which defines the
-- configuration variables <root_directory>, <filename_filter>,
-- <lua_diretory>, <html_diretory> and <users>, all of them fields
-- of <cgilua>, finds the user script to be executed and changes
-- the process uid. this file is executed without some global symbols from iolib
-- for security reasons; the symbols are restored only after the process uid
-- is changed to the scripts's owner uid.
-- this function defines <cgilua.script_path> as the physical path
-- of the user script to be executed

function domap()
  local ref = restrict( cgilua.restore_external, domap )
  cgilua.remove_external()
  domap = nil
  cgilua.restore_external = nil
  cgilua.dofile( "map.lua" )
  cgilua.script_path = cgilua.getuserscript()
  -- get rid of root permissions
  owneruid( cgilua.script_path )
  cgilua.restore_external = getrestricted( ref )
  cgilua.restore_external()
end

domap()


-- the function <splitpath> takes a path and returns its absolute
-- directory and file parts.

function cgilua.splitpath( path )
  local i, f, dir, file = strfind( path, "(.*[/\\])([^/\\]*)$" )
  if not i then
    return getcwd()..'/', path
  elseif not strfind( '/:\\', strsub(dir, 1, 1)) and
         not strfind( strsub(dir, 1, 2), '%a:')  then
    return getcwd()..'/'..dir, path
  end
  return dir, file
end
cgilua.split = cgilua.splitpath


-- redefines <cgilua.directory> to use an absolute path
cgilua.directory = cgilua.splitpath( cgilua.conffile )



-- the function <cgilua.doscript> process a script file,
-- executing or preprocessing it.

function cgilua.doscript( script )
  if not strfind( gsub( script, ".*/([^/]*)", "%1" ),
                        cgilua.filename_filter ) then
    if not cgilua.douserfile( script ) then
      cgilua.error( format( "Error reading lua file <code>%s</code>.", script ) )
    end
  else
    if cgilua.preprocess then
      cgilua.preprocess( script )
    else
      if not readfrom( script ) then
        cgilua.error( format( "Error reading html file <code>%s</code>.", script ) ) 
      else
        write( read( ".*" ) )
        readfrom()
      end
    end
  end
end


-- the function <cgilua.run> 
-- define the variables <cgilua.cgilua_url>, <cgilua.script_path>,
-- <cgilua.script_vdir> and <cgilua.script_pdir>;
-- change the current directory to the script directory and do
-- (execute or preprocess) the script.

function cgilua.run()

  if cgilua.param then

    local script_name = getenv( "SCRIPT_NAME" )
    cgilua.cgilua_url = ''
    if script_name then
      cgilua.script_vdir = cgilua.splitpath( script_name )
    end

  else

    cgilua.script_vdir  = cgilua.splitpath( getenv("PATH_INFO") )
    cgilua.cgilua_url   = getenv( "SCRIPT_NAME" )

  end

  -- extracts the directory part of the path
  cgilua.script_pdir  = cgilua.splitpath( cgilua.script_path )

  -- change directory to where the lua file resides
  chdir( cgilua.script_pdir )

  -- do (execute ou preprocess) the script file
  cgilua.doscript( cgilua.script_path )
end


-- the function <cgilua.relative_url> returns the url
-- to be used to activate a script in the same directory
-- of the current being executed.
-- this prevents the coding of fixed paths in the scripts,
-- keeping the scripts independent from the directory it is installed
-- (and from the directory cgilua is installed).

function cgilua.relative_url( script )
  return cgilua.cgilua_url..cgilua.script_vdir..script
end


-- the global variable <cgilua.stdin> holds all the
-- input submited via standard input.

cgilua.stdin = nil


-- the following code performs the form data decoding:

-- the table <hexa> maps hexadecimal digits
-- to its equivalent number. it is used
-- to map a string with a hexadecimal number
-- to a number.

cgilua.hexa = { h0=0,  h1=1,  h2=2,  h3=3,
                h4=4,  h5=5,  h6=6,  h7=7,
                h8=8,  h9=9,  hA=10, hB=11,
                hC=12, hD=13, hE=14, hF=15 }


-- the function <hexa2char> takes two hexadecimal digits
-- as strings and returns the correnpondent 
-- character with decimal code <ms>*16+<ls>.

function cgilua.hexa2char( ms, ls )
  return format( "%c", cgilua.hexa['h'..ms]*16 + cgilua.hexa['h'..ls] )
end


-- the function <unescape> does the decoding of the
-- submited data, replacing every occurence of the
-- substring %XX (where XX is a hexadecimal number)
-- with the equivalent character of decimal code XX

function cgilua.unescape( str )
  str = gsub( str, "+", " " )
  str = gsub( str, "%%0D%%0A", "\n" )
  return gsub( str, "%%([0-9A-F])([0-9A-F])", cgilua.hexa2char )
end

-- the function <pair> inserts the pair (key,value) on
-- cgilua's main table after decoding the value of the components.  

function cgilua.pair( key, value )
  cgilua.insertfield( cgilua.unescape(key), cgilua.unescape(value) )
end


-- the function <parse> parses the form data.
-- it takes the encoded data from the standard input
-- or from the environment variable QUERY_STRING,
-- depending on the form method (POST or GET).
-- it transforms the '+' character into a space (' '),
-- a sequence '%0D%0A' into an linefeed ('\n')
-- and calls the function <pair> for each form element encoded.

function cgilua.parse()
  local input
  if getenv('REQUEST_METHOD') == 'POST' then
    local content_lenght = getenv("CONTENT_LENGTH")
    input = read( strrep( '.', content_lenght) ) 
    cgilua.stdin = input
  else
    input = getenv("QUERY_STRING") 
    if not input then return end
  end
  gsub( input, "&*([^&=]*)=([^&=]*)&*", cgilua.pair )
end

-- the function <char2hexa> converts a character
-- into a hexadecimal representation of its code number,
-- returning a string of the form %XX

function cgilua.char2hexa( c )
  return '%'..format( "%02X", ascii(c) )
end

-- the function <escape> does the encoding of the
-- string given by parameter, replacing every occurence 
-- of a character not in the range a-z, A-Z and 0-9
-- with a string on the form %XX, where XX is
-- the hexadecimal representation of the character's
-- number code.

function cgilua.escape( s )
  s = gsub( s, "([^a-zA-Z0-9 \n])", cgilua.char2hexa )
  s = gsub( s, " ", "+" )
  s = gsub( s, "\n", "%%0D%%0A" )
  return s
end

-- the function <encodecgi> returns a string with all 
-- the elements of the table cgi encoded in a way
-- adequate to be used with the GET method.  

function cgilua.encodecgi()
  local str = ''
  local key,value = next( cgi, nil )

  while key do
    str = str..cgilua.escape( key )..'='..cgilua.escape( value )..'&' 
    key,value = next( cgi, key )
  end

  -- remove last &
  return strsub( str, 1, strlen(str)-1 )
end


-- the function <cgilua.includehtml> outputs the contents
-- of the file specified as an argument.

function cgilua.includehtml( file )
  if not readfrom( file ) then
    cgilua.error( format( 'Error opening file <CODE>%s</CODE>. '..
                  'File cannot be included.', file ) )
  else
    cgilua.processtext( read(".*") )
    readfrom()
  end
end


-- parse the form data

cgilua.parse()


-- define the error treatment for user scripts

cgilua.dofile( 'error.lua' )


-- define the lua environment for the user script 

cgilua.dofile( 'env.lua')


-- get and 'do' the lua file or preprocess the text file!

cgilua.run()


