$debug


-- <cgilua.showstacklevel> prints one level of the stack
-- described in <l>. It prints the normal Lua execution stack
-- treating a new type 'embedded', defined for the chunks
-- on code embedded in html files.

function cgilua.showstacklevel( l )
  if l.type == 'global' then
    print( format( "function <code>%s</code>", l.name ) )
    if l.linedef > 0 then
      if strsub(l.file, 1, 10) ~= '(dostring)' then
        print( format( " (<i>%s</i>:%d)", l.file, l.linedef) )
      else
        print( " (<i>dostring or html file</i>)" )
      end
    end
  elseif l.type == 'fallback' then
    print( format( "<code>%s</code> fallback", l.name ) )
  elseif l.type == 'embedded' then
    print( format( "embedded code <i>%s</i>", l.name ) )
  else
    if l.linedef == 0 then
      print( format( "main of <i>%s</i>", l.file) )
    elseif l.linedef < 0 then
      print( format( "<i>%s</i>", l.file) )
    else
      print( format( "function (<i>%s</i>:%d)", l.file, l.linedef) )
    end
  end
  if l.line > 0 then
    print( format( " at line %d", l.line) )
  end
  print("<br>\n");
end


-- this is the function called by the error fallback
-- to show the Lua stack. It manipulates the stack
-- to generate a stack adequate to CGILua, hiding some
-- internal processing and showing in the stack
-- embedded code, html files etc.
-- It searches the stack for the functions PRE_ProcessFile, cgilua.usererror,
-- cgilua.run, cgilua.doscript, cgilua.douserfile and cgilua.douserstring.

function cgilua.printstack( stack )
  local i = 1
  local j = 1
  local toprint = {}

  while stack[i] do

    -- PRE_ProcessFile:
    -- remove it from the stack and put the name of the
    -- html file being processed in the same way as a lua file
    -- executed with 'dofile'
    if stack[i].func == PRE_ProcessFile then

      local str = cgilua.pop( 'script' )
      toprint[j] = {}
      toprint[j].type = ''
      toprint[j].line = 0
      toprint[j].linedef = 0
      toprint[j].file = str
      i = i + 1
      j = j + 1

    -- cgilua.doscript:
    -- shows the name of the this function, instead of showing the
    -- file name and line where it was defined.
    elseif stack[i].func == cgilua.doscript then

      toprint[j] = stack[i]
      toprint[j].type = 'global'
      toprint[j].name = 'cgilua.doscript'
      i = i + 1
      j = j + 1

    -- cgilua.run:
    -- this is the base of the stack.
    -- it calls cgilua.doscript, which is removed from the stack
    -- this function and the following stack levels are also removed
    elseif stack[i].func == cgilua.run then

      toprint[j-1] = nil
      stack[i] = nil

    -- cgilua.douserfile:
    -- remove it from the stack ('dofile' is enough)
    elseif stack[i].func == cgilua.douserfile then

      j = j - 1
      toprint[j] = nil
      i = i + 1

    -- cgilua.usererror:
    -- remove it and the pre-process functions following this stack level
    elseif stack[i].func == cgilua.usererror then
    
      j = j - 1
      toprint[j] = nil
      i = i + 1
      while stack[i] and stack[i].func ~= PRE_ProcessFile and
            stack[i].name and strsub(stack[i].name, 1, 4) == 'PRE_' do
        i = i + 1
      end

    -- cgilua.douserstring:
    -- shows the dostring as an embedded code and remove
    -- the pre-process functions following this stack level
    elseif stack[i].func == cgilua.douserstring then

      local str = cgilua.pop( 'embedded' )

      j = j - 1
      toprint[j] = nil

      str = gsub( str, "%s%s*", " " );
      if strlen(str) > 50 then
        str = strsub( str, 1, 50 ) .. ' ...'
      end
      str = gsub( str, "<", "&lt;" )
      str = gsub( str, " ", "&nbsp;" )

      toprint[j-1].name = str
      toprint[j-1].type = 'embedded'
      i = i + 1
      while stack[i] and stack[i].func ~= PRE_ProcessFile and
            stack[i].name and strsub(stack[i].name, 1, 4) == 'PRE_' do
        i = i + 1
      end

    else

      toprint[j] = stack[i]
      i = i + 1
      j = j + 1

    end    
  end
 
  -- print the manipulated stack
  i = 1
  while toprint[i] do
    cgilua.showstacklevel( toprint[i] )
    i = i + 1
  end

end

-- the functions <cgilua.{create_stack, push, pop}>
-- are used to manage the cgilua stacks

cgilua.stacks = {}
cgilua.stacks.tops = {}

function cgilua.create_stack( name )
  cgilua.stacks[name] = {}
  cgilua.stacks.tops[name] = 0
end

function cgilua.push( stack, value )
  local top = cgilua.stacks.tops[stack]
  cgilua.stacks[stack][ top ] = value
  cgilua.stacks.tops[stack] = top + 1
end

function cgilua.pop( stack )
  local top = cgilua.stacks.tops[stack] - 1
  cgilua.stacks.tops[stack] = top 
  return cgilua.stacks[stack][ top ]
end


-- create two stacks

cgilua.create_stack( 'script' )
cgilua.create_stack( 'embedded' )


-- define wrappers for the functions dofile and dostring,
-- setting the user_fallback before the call to dofile (dostring)

function cgilua.douserstring( luacode, embedded )
  local old
  cgilua.push( 'embedded', embedded )
  old = setfallback( 'error', cgilua.userfallback );
  result = dostring( luacode )
  setfallback( 'error', old );
  cgilua.pop( 'embedded' )
  return result
end

function cgilua.douserfile( file )
  local old = setfallback( 'error', cgilua.userfallback )  
  local result = dofile( file )
  setfallback( 'error', old )
  return result
end

function cgilua.usererror( msg )
  setfallback( 'error', cgilua.userfallback )  
  error( msg )
end

