$debug

--=============================================================================

-- This file implements a toolkit for the design of HTML pages. The mainly
-- concern is to create a compact and homogeneous way where HTML tags can 
-- be built, using the power of the lua language.
-- Tags cad

--=============================================================================


------------------------------------------------------------------------------

-- Inherit
-- This function implements simple inheritance with
-- lua's fallback mecanism.

function Inherit (object,field)
  if field == "parent" then
    return OldIndex(t,f)
  end
  local p = object.parent
  if type(p) == 'table' then
    return p[field]
  else
    return OldIndex(object,field)
  end
end

OldIndex = setfallback("index",Inherit)

-------------------------------------------------------------------------------

-- strsplit
-- This function splits a string according to a delimiter given.
-- It is returned the array with th splitted strings and the number of
-- substrings produced

function strsplit( str, delim )
  local pos = 1
  local oldpos = 1
  local array = {}
  local i = 1

  pos = strfind( str, delim, oldpos, 1 )
  if pos then
    while pos do
      array[i] = strsub( str, oldpos, pos-1 )
      oldpos = pos + strlen( delim )
      i = i + 1
      pos = strfind( str, delim, oldpos, 1 )

      if pos == nil then
        array[i] = strsub( str, oldpos )
      end
    end
  else
    array[i] = str
  end

  return array,i
end

-------------------------------------------------------------------------------

-- Main context 

HtmlTag = {}


-------------------------------------------------------------------------------

-- HtmlTag:DeclareTag
-- This method puts the calling object on the hierarchy

function HtmlTag:DeclareTag (o) 
  o.parent = self 
  return o 
end


-------------------------------------------------------------------------------

-- HtmlTag:SetContainer
-- This method move the objects indexed in the tag element with 
-- numbers to the 'contain' field.

function HtmlTag:SetContainer()
  local field,value = next ( self, nil )

  while field do
    if type ( field ) == 'number' then
      if not self.contain then
        self.contain = {}
      end
      self.contain [ field ] = value
      self [ field ] = nil
    end
    field, value = next( self, field )
  end
end


-------------------------------------------------------------------------------

-- HtmlTag:SetFields
-- This method performs the checking of the fields of the element
-- newly created. The functions to be applied on the field and the
-- fields to be checked are specified on the 'fields' field of
-- its parent Descriptor.

function HtmlTag:SetFields()
  local attrib, f = next ( self.parent.fields, nil )

  while attrib do
    -- if the field has string type, call its check function
    -- f[1] is the check function and f[2] its opcional parameters
    if type ( attrib ) == 'string' then
      self [ attrib ] = f[1] ( self, attrib , f[2] ) -- f[1] is a method
    end

    attrib, f = next( self.parent.fields, attrib )
  end
end


-------------------------------------------------------------------------------

-- HtmlTag:BuildElement
-- This method is called by every tag's constructor and it is
-- responsible for applying the methods SetContainer and SetFields.

function HtmlTag:BuildElement(t)
  t.parent = self
  t:SetContainer()
  t:SetFields()
  return t
end


-------------------------------------------------------------------------------

-- HtmlTag:Print
-- This method prints the element (and its child elements) on the 
-- HTML format.

function HtmlTag:Print()
  local attrib, value = next ( self , nil )
  local i = 1

-- write('\n')
  -- first print the attributes of the element itself
  if self.tag ~= '' then
    write('<'.. self.tag)
    while attrib do
      if type (attrib) == 'string' and attrib ~= 'parent' and attrib ~= 'contain'  then
        write(' '..attrib..'="'..value..'"')
      end
      attrib, value = next( self, attrib )
    end
    write('>')
  end

  -- then go recursively on the child elements
  if self.contain then   
    while self.contain[i] do
      value = self.contain[i]
      if type(value) == 'table' and value.tag then
        value:Print()
      elseif type(value) == 'string' then
        write(value)
      end
      i = i + 1
    end

    if self.tag ~= '' then
      write('</'.. self.tag..'>')
    end
  end
end


-------------------------------------------------------------------------------

-- HtmlTag:Error
-- This method calls lua's error function to send errors messages

function HtmlTag:Error( str )
  error ( 'Error on Html tag. '..str )
end


-------------------------------------------------------------------------------

-- HtmlTag:AttribRequired
-- This method is called for checking fields that must be present
-- in the element newly created.

function HtmlTag:AttribRequired ( a, attrib_type )
  if not self[a] then
    self:Error ('Attribute "'..a..'" is required for <'.. self.tag ..'> tag.')
  end
  if attrib_type then
    if type ( self[a] ) ~= attrib_type then
      self.Error ('Attribute '..a..' of <'..self.tag..'> tag must be a '..
                  attrib_type..' type. ')
    end
  end
  return self[a]
end


-------------------------------------------------------------------------------

-- HtmlTag:ApplyOnFields
-- This method applies modifications the the fields specified,
-- acting recursively on the child elements.

function HtmlTag:ApplyOnFields( string_of_fields, what_to_apply, func_param  )
  local fields
  local component, value 
  local result
  local i = 1

  if type( string_of_fields ) == 'string' then
    fields = strsplit( string_of_fields, ",")
  else
    fields = string_of_fields
  end

  while fields[i] do

    if self[fields[i]] then
      if type( what_to_apply ) == 'string' then

        -- case of replacing one field with another
        if self[what_to_apply] then
          self[fields[i]] = self[what_to_apply]
          self[what_to_apply] = nil
        else
        -- new value for the field
          self[fields[i]] = what_to_apply
        end

      -- second parameter is a function
      elseif type( what_to_apply ) == 'function' then
        self[fields[i]] = what_to_apply( func_param )
      end
    end

    i = i + 1
  end

  if self.contain then  
    component, value = next( self.contain, nil )

    while component do
      if type(value) == 'table' and value.tag then
        value:ApplyOnFields( fields, what_to_apply, func_param  )
      end

      component, value = next( self.contain, component )
    end
  end
end


-------------------------------------------------------------------------------

IMG_Descriptor = HtmlTag:DeclareTag {
  tag = "img",
  fields = {
    src = { HtmlTag.AttribRequired , 'string' }
  }
}

B_Descriptor = HtmlTag:DeclareTag {
  tag = "b",
  fields = {
  }
}

P_Descriptor = HtmlTag:DeclareTag {
  tag = "p",
  fields = {
  }
}

A_Descriptor = HtmlTag:DeclareTag {
  tag = "a",
  fields = {
  }
}

BR_Descriptor = HtmlTag:DeclareTag {
  tag = "br",
  fields = {
  }
}

BOX_Descriptor = HtmlTag:DeclareTag {
  tag = "",
  fields = {
  }
}


-------------------------------------------------------------------------------

function IMG (param)
  return IMG_Descriptor:BuildElement(param)
end

function B (param)
  return B_Descriptor:BuildElement(param)
end

function P (param)
  return P_Descriptor:BuildElement(param)
end

function A (param)
  return A_Descriptor:BuildElement(param)
end

function BR (param)
  return BR_Descriptor:BuildElement(param)
end

function BOX (param)
  return BOX_Descriptor:BuildElement(param)
end

-------------------------------------------------------------------------------
