$debug

--=============================================================================

-- This file implements the preprocessing of a text file. Text not marked
-- to be processed is simply repeated on the output. There are three
-- kinds of marks that can be used in order to activate the processing:
--
-- (1) <!--$$ lua_code $$-->
--     where "lua_code" is a valid lua construction. This lua code is executed,
--     and if it contains expressions that write to the stdout, this
--     happens exactly at the point of the file where it is positioned.
--     Note that according to HTML syntax, everything enclosed by the
--     marks "<!--" and  "-->" is not displayed, so the lua code
--     is not visible on a browser if the file is viewed without
--     the preprocessing.
--
-- (2) $| lua_expression |$
--     where "lua_expression" is a valid lua expression. This lua 
--     expression is evaluated, and its result is printed. If the result 
--     is an element made with the "htmltoolkit" package, the element is 
--     recognized and properly printed. 
--
-- (3) <!--$$LOOP start='cmd_start', test='cond', action='cmd_action' $$-->
--       text
--     <!--$$ENDLOOP $$-->
--     where "cmd_start" is a lua code executed before the loop starts;
--     "cond" is a condition tested for the continuation of the loop,
--     if it is evaluated as false the loop stops; 
--     "cmd_action" is a lua code executed at the end of every loop cycle; and
--     "text" is a generic chunk of text that may contain any of the
--     three types of marks and that is continually submited for processing
--     until the loop stops.
--
--

--=============================================================================

-- PRE_ProcessFile
-- This is the main function of the preprocessing package and the only one
-- that should be called outside this file. It receives an argument, the 
-- name of the file to be processed, and if there is no error opening
-- the file, the processing is started.

function PRE_ProcessFile( htmlfile )

  if readfrom( htmlfile ) then
    write( 'Content-type: text/html\n\n' )
    cgilua.push( 'script', htmlfile )
    PRE_ProcessText( read( ".*" ) )
    cgilua.pop( 'script' )
    readfrom()
    return 1
  else
    cgilua.error( format( 'Error opening file <CODE>%s</CODE>. '..
                  'Preprocessing aborted.', htmlfile ) )
  end

  return nil
end

cgilua.preprocess = PRE_ProcessFile


------------------------------------------------------------------------------


-- PRE_ReplaceAndPrint
-- This function searches for text chunks on the form "$|EXP|$" and
-- replaces them with the result of the lua expressions EXP.
-- If the expression is an element produced by the "htmltoolkit",
-- then the element itself knows how to be represented (the method
-- "Print" is called).

function PRE_ReplaceAndPrint( chunk )
  local s, e, expr, repl, old  

  s,e,expr = strfind( chunk, "%$%|([^$]*)%|%$" )

  -- a match has been found in the text chunk?
  while s do
    -- write the text that need not to be analysed
    write( strsub( chunk,1,s-1) )

    -- evaluates the lua expression and print it
    repl = cgilua.douserstring( 'return '..expr, '$|'..expr..'|$' ) 
    if type( repl ) == 'table' and repl.tag then
      repl:Print()
    elseif type( repl ) == 'number' or type( repl ) == 'string' then
      write( repl )
    elseif repl ~= nil then
      cgilua.error( 'Error processing the expression '..
                    '<CODE>$|'..expr..'|$</CODE>. \nThe resulting '..
                    'item (type <CODE>'..type(repl)..'</CODE>) '..
                    'cannot be printed.')
    end

    -- continue to search for expressions to be replaced
    -- on the rest of the text chunk
    chunk = strsub( chunk, e+1)
    s,e,expr = strfind( chunk,"%$%|([^$]*)%|%$")
  end

  -- write the final part of the text chunk
  write( chunk )
end


------------------------------------------------------------------------------

-- PRE_FindEndloop
-- This function is used on PRE_ExecuteLoop to determine the limits where
-- the LOOP command should be applied. This corresponds to finding the
-- outmost ENDLOOP statement. 

function PRE_FindEndloop( chunk )
  local count=1
  local match, skip, prefix, pointer, endloop, out

  pointer = chunk
  out = 0

  repeat
    -- searches for the first occurence of ENDLOOP or LOOP
    match, skip, prefix = strfind( pointer,"<!--%$%$%s*(%a*)LOOP[^$]*%$%$-->\n?")
    -- if there is not a match, abort. There must be an closing ENDLOOP.
    if not match then return 0 end

    -- increments the counter in the case of a LOOP, or
    -- decrements it in the case of ENDLOOP
    endloop = strfind( prefix, "END", 1, 1)
    if endloop then
      count = count - 1
    else
      count = count + 1
    end

    -- go ahead with the search
    pointer = strsub( pointer, skip+1 )

    -- update the distance from the start of the text chunk
    out = out + skip 
  until count == 0 

  -- return the position on the chunk just before the matching ENDLOOP
  return out-(skip-match+1)
end


------------------------------------------------------------------------------

-- PRE_ExecuteLoop
-- This function executes a loop, using the control parameters specified
-- on the LOOP statement. The first argument is the chunk of text where
-- the corresponding ENDLOOP should be found. The second argument is
-- the control parameters obtained from the LOOP statement. 
-- A LOOP statement have the form:
-- <!--$$LOOP start='cmd', test='cond', action='cmd2' $$-->
-- Note that there must be 3 control parameters, separated
-- by commas: "start", "test" and "action". The value of this parameters
-- must be a string of a valid lua expression. 
-- "start" is a command executed one time before the loop starts;
-- "test" is a condition tested for the continuation of the loop; and
-- "action" is a command executed at the end of every loop cycle.

function PRE_ExecuteLoop ( text, param )
  local control,endchunk,chunk

  -- convert the control parameters into fields of a lua table
  control = cgilua.douserstring( 'return {'..param..'}', '<!--$$LOOP'..param..'$$-->' )
  if not control.start or not control.action or not control.test then
    cgilua.usererror( "incorrect loop statement" )
  end

  -- finds the position where the loop ends
  endchunk = PRE_FindEndloop( text )

  -- separate the chunk of text where the loop act
  chunk = strsub( text, 1, endchunk )

  -- do the "start" control statement
  cgilua.douserstring(control.start, 'start = '..control.start )

  -- while the "test" control statement is valid, run all over again
  -- the main PRE_ProcessText function, building a stack of chunks
  -- being processed. At the end, the "action" control statement
  -- is executed
  while cgilua.douserstring('return '..control.test, 'test = '..control.test ) do
    PRE_ProcessText( chunk )
    cgilua.douserstring(control.action, 'action = '..control.action )
  end

  -- returns the text after then closing ENDLOOP
  return strsub( text, endchunk+1 )
end


------------------------------------------------------------------------------

-- PRE_ProcessText
-- This function coordinates the action of the others and executes the
-- main loop of the processing: it goes through the text searching
-- for special marks. There are two types of special marks:
-- 1) The pair  "<!--$$" and "$$-->";      
-- 2) The pair  "$|" and "|$".
-- On the first type all the text enclosed is captured and executed as
-- lua code. On the second type the text enclosed is considered a lua
-- expression and an attempt is made to print it. 

function PRE_ProcessText( text )
  local head, tail, cmd, chunk, loop, skip,match

  head,tail,cmd = strfind(text,"<!--%$%$([^$]*)%$%$-->\n?")

  while head do

    -- go back and try to find expressions of the form $|EXP|$
    chunk = strsub (text, 1, head-1)
    PRE_ReplaceAndPrint( chunk )

    -- test if the <!--$$ cmd $$--> found is a LOOP statement
    loop,skip = strfind( cmd, "LOOP", 1, 1)

    text  = strsub ( text, tail+1)
    if loop then
      -- execute the LOOP statement
      text = PRE_ExecuteLoop( text, strsub ( cmd, skip+1 ) )

      -- after executing the LOOP statement, skip the matching ENDLOOP
      match, skip = strfind( text,"<!--%$%$%s*ENDLOOP[^$]*%$%$-->\n?")
      if match then
        text =  strsub( text, skip+1 )
      else
        cgilua.error( 'Error processing the file. \n'..
                      'The matching <CODE><!--$$ENDLOOP $$--></CODE> '..
                      'statement could not be found.' )
      end
    else
      -- execute the lua code enclosed on <!--$$ ... $$-->
      cgilua.douserstring(cmd, '<!--$$ '..cmd..' $$-->' )
    end
    -- continue to search
    head,tail,cmd = strfind( text, "<!--%$%$([^$]*)%$%$-->\n?")
  end

  -- search for $| ... |$ on the final part of the text
  PRE_ReplaceAndPrint( text )
end

cgilua.processtext = PRE_ProcessText

